SQLITE_URI = 'sqlite:///db.sqlite'
POSTGRES_URI = 'postgresql://postgres:postgres@localhost:5432/postgres'
MYSQL_URI = 'mysql://mysql:mysql@localhost:3306/mysql'
REDIS_URI = 'redis://localhost:6379/0'
